/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.notification.base;

import org.apache.ws.addressing.EndpointReference;
import org.apache.ws.notification.topics.expression.TopicExpression;
import org.apache.ws.pubsub.NotificationConsumer;
import org.apache.ws.pubsub.NotificationProducer;
import org.apache.ws.resource.PropertiesResource;
import org.apache.ws.resource.lifetime.ScheduledResourceTerminationResource;
import org.apache.ws.resource.properties.query.QueryExpression;

/**
 * Interface to be implemented by subscription resources. Exposes the information
 * associated with a subscription request, as well as resource lifetime-related
 * state.
 */
public interface Subscription extends PropertiesResource, org.apache.ws.pubsub.Subscription, ScheduledResourceTerminationResource
{

    /**
     * Gets the topic expression associated with this subscription.
     *
     * @return the topic expression
     */
    TopicExpression getTopicExpression();

    /**
     * Gets the precondition associated with this subscription.
     *
     * @return the precondition, or null if no precondition was specified in the subscribe request
     */
    QueryExpression getPrecondition();

    /**
     * Gets the selector associated with this subscription.
     *
     * @return the selector, or null if no selector was specified in the subscribe request
     */
    QueryExpression getSelector();

    /**
     * Wrap notification messages in the notify element?
     *
     * @return true (default) if notify should be used, or false if not
     */
    boolean getUseNotify();

    /**
     * Returns true if this subscription is currently paused, or false otherwise.
     *
     * @return true if this subscription is currently paused, or false otherwise
     */
    boolean isPaused();

    /**
     * Temporarily suspends the delivery of notification messages for this subscription.
     *
     * @throws Exception if unable to suspend message delivery
     */
    void pause() throws Exception;

    /**
     * Resumes the delivery of notification messages for this subscription.
     *
     * @throws Exception if unable to resume message delivery
     */
    void resume() throws Exception;

    /**
     * Gets the producer resource associated with this subscription.
     *
     * @return the producer resource
     */
    NotificationProducerResource getProducerResource();

    //TODO: remove this method once getEPR() method is added to Resource interface
    void setEndpointReference(EndpointReference epr);

    // TODO: figure out if we need this...
    void setNotificationConsumer(NotificationConsumer notificationConsumer);

    // TODO: figure out if we need this...
    void setNotificationProducer(NotificationProducer notificationProducer);

    /**
     * Gets the consumer EPR associated with this subscription.
     *
     * @return the consumer EPR associated with this subscription
     */
    EndpointReference getConsumerReference();

    void setPolicy( Object policy );

    void setPrecondition( QueryExpression precondition );

    void setSelector( QueryExpression selector );

    void setUseNotify( boolean useNotify );    

}
