/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.notification.base.impl;

import org.apache.ws.addressing.EndpointReference;
import org.apache.ws.notification.base.NotificationProducerResource;
import org.apache.ws.notification.base.Subscription;
import org.apache.ws.notification.topics.Topic;
import org.apache.ws.notification.topics.TopicListener;
import org.apache.ws.notification.topics.TopicListenerList;
import org.apache.ws.notification.topics.TopicSpaceSet;
import org.apache.ws.notification.topics.expression.TopicExpression;
import org.apache.ws.notification.topics.expression.TopicExpressionException;
import org.apache.ws.notification.topics.impl.SubscriptionTopicListener;
import org.apache.ws.pubsub.Filter;
import org.apache.ws.pubsub.NotificationConsumer;
import org.apache.ws.pubsub.NotificationProducer;
import org.apache.ws.resource.PersistenceCallback;
import org.apache.ws.resource.PropertiesResource;
import org.apache.ws.resource.Resource;
import org.apache.ws.resource.ResourceException;
import org.apache.ws.resource.ResourceHome;
import org.apache.ws.resource.ResourceKey;
import org.apache.ws.resource.lifetime.ResourceTerminationEvent;
import org.apache.ws.resource.lifetime.ResourceTerminationListener;
import org.apache.ws.resource.lifetime.impl.ResourceTerminationEventImpl;
import org.apache.ws.resource.properties.ResourcePropertySet;
import org.apache.ws.resource.properties.query.QueryExpression;
import org.apache.ws.util.uuid.UuidGenerator;
import org.apache.ws.util.uuid.UuidGeneratorFactory;

import javax.naming.InitialContext;
import javax.xml.rpc.JAXRPCException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;

/**
 * TODO
 */
public abstract class AbstractSubscription implements Subscription, PropertiesResource
{

    //private static final Log LOG = LogFactory.getLog( AbstractSubscription.class.getName() );

    protected EndpointReference m_consumerReference;
    protected EndpointReference m_producerReference;
    protected Object m_policy;
    protected QueryExpression m_precondition;
    protected QueryExpression m_selector;
    protected ResourceKey m_producerKey;
    protected String m_producerHomeLocation;
    protected TopicExpression m_topicExpression;
    protected boolean m_isPaused;
    protected boolean m_useNotify = true;
    protected Calendar m_terminationTime;
    protected Calendar m_creationTime;
    protected String m_id = null;

    protected EndpointReference m_epr;

    transient protected ResourcePropertySet m_propSet;

    private static final UuidGenerator UUID_GEN =
            UuidGeneratorFactory.createUUIdGenerator();
    private NotificationConsumer m_notificationConsumer;
    private NotificationProducer m_notificationProducer;

    /**
     * A list of termination listeners to be notified when the resource is terminated.
     */
    private List m_terminationListeners = new ArrayList();

    /**
     * Construct a new subscription resource.
     *
     * @param consumerReference      The WS-Addressing endpoint reference of the consumer
     * @param producerReference      The WS-Addressing endpoint reference of the producer
     * @param producerKey            The key of the producer resource
     * @param producerHomeLocation   The JNDI location of the home of the producer resource
     * @param topicExpression        The topic expression for this subscription
     */
    public AbstractSubscription( EndpointReference consumerReference,
                                 EndpointReference producerReference,
                                 ResourceKey producerKey,
                                 String producerHomeLocation,
                                 TopicExpression topicExpression )
    {
        m_id = UUID_GEN.generateUuid();
        m_consumerReference = consumerReference;
        m_producerReference = producerReference;
        m_producerKey = producerKey;
        m_producerHomeLocation = producerHomeLocation;
        m_topicExpression = topicExpression;
        m_creationTime = Calendar.getInstance();
    }

    public ResourcePropertySet getResourcePropertySet()
    {
        return m_propSet;
    }

    public void setTerminationTime( Calendar time )
    {
        m_terminationTime = time;
    }

    public Calendar getTerminationTime()
    {
        return m_terminationTime;
    }

    public Calendar getCurrentTime()
    {
        return Calendar.getInstance();
    }

    public Calendar getCreationTime()
    {
        return m_creationTime;
    }

    public EndpointReference getConsumerReference()
    {
        return m_consumerReference;
    }

    public Object getPolicy()
    {
        return m_policy;
    }

    public QueryExpression getPrecondition()
    {
        return m_precondition;
    }

    public EndpointReference getProducerReference()
    {
        return m_producerReference;
    }

    public NotificationProducerResource getProducerResource()
    {
        try
        {
            ResourceHome producerHome = (ResourceHome) new InitialContext().lookup( m_producerHomeLocation );
            return (NotificationProducerResource) producerHome.find( m_producerKey );
        }
        catch ( Exception e )
        {
            throw new RuntimeException( "Failed to lookup NotificationProducer resource due to internal error: " + e, e );
        }
    }

    public QueryExpression getSelector()
    {
        return m_selector;
    }

    public TopicExpression getTopicExpression()
    {
        return m_topicExpression;
    }

    public boolean isPaused()
    {
        return m_isPaused;
    }

    public void pause() throws Exception
    {
        m_isPaused = true;
    }

    public void resume() throws Exception
    {
        m_isPaused = false;
    }

    public boolean getUseNotify()
    {
        return m_useNotify;
    }

    public AbstractSubscription()
    {

    }

    public Object getID()
    {
        return m_id;
    }

    public void setID( Object o )
    {
        m_id = (String) o;
    }

    public void destroy()
    {
        Resource producerResource = null;
        try
        {
            producerResource = getProducerResource();
        }
        catch ( Exception e )
        {
            ;
        }

        if ( !( producerResource instanceof NotificationProducerResource ) )
        {
            return;
        }

        //todo fix!
        //TopicSpaceSet topicSpaceSet = ( (NotificationProducerResource) producerResource ).getTopicSpaceSet();
        Topic[] topics = null;
        try
        {
            //TopicSpace topicSpace = topicSpaceSet.getTopicSpace();
            //topics = topicSpace.getTopics( m_topicExpression );
            topics = evaluateTopicExpression();

        }
        catch ( Exception e )
        {
            ;
        }

        synchronized ( producerResource )
        {
            //removes listeners from the topics
            boolean removed = removeListener( topics );

            //persists the resource
            if ( removed && producerResource instanceof PersistenceCallback )
            {
                try
                {
                    ( (PersistenceCallback) producerResource ).store();
                }
                catch ( ResourceException e )
                {
                    ;
                }
            }
        }

        //notify listeners that this subscription expired.
        ResourceTerminationEvent rte = new ResourceTerminationEventImpl( getID(), "Subscription Destroyed" );
        for ( int i = 0; i < m_terminationListeners.size(); i++ )
        {
            ResourceTerminationListener resourceTerminationListener = (ResourceTerminationListener) m_terminationListeners.get(
                    i );
            resourceTerminationListener.terminationOccurred( rte );
        }

    }

    private Topic[] evaluateTopicExpression() throws Exception
    {
        NotificationProducerResource resource = (NotificationProducerResource) getProducerResource();
        TopicSpaceSet topicSpaceSet = resource.getTopicSpaceSet();
        try
        {
            return topicSpaceSet.evaluateTopicExpression( m_topicExpression );
        }
        catch ( TopicExpressionException e )
        {
            throw new JAXRPCException( "An exception occurred during subscription. ", e );
        }
    }

    private boolean removeListener( Topic[] topics )
    {
        boolean removed = false;
        for ( int i = 0; i < topics.length; i++ )
        {
            TopicListenerList topicListenerList = topics[i];
            synchronized ( topicListenerList )
            {
                Iterator topicListenerIterator = topicListenerList.topicListenerIterator();
                while ( topicListenerIterator.hasNext() )
                {
                    TopicListener listener = (TopicListener) topicListenerIterator.next();
                    if ( listener instanceof SubscriptionTopicListener )
                    {
                        SubscriptionTopicListener lt =
                                (SubscriptionTopicListener) listener;
                        Object subKey =
                                lt.getSubscription().getID();
                        if ( subKey.equals( m_id ) )
                        {
                            topicListenerIterator.remove();
                            removed = true;
                        }
                    }
                }
            }
        }

        return removed;
    }

    public void unsubscribe()
    {
        destroy();
    }

    public Filter getFilters()
    {
        // TODO. remove?
        return null;
    }

    public URI getDeliveryMode()
    {
        // TODO. remove?
        return null;
    }

    public NotificationConsumer getNotificationConsumer()
    {
        return m_notificationConsumer;
    }

    public NotificationProducer getNotificationProducer()
    {
        return m_notificationProducer;
    }

    public EndpointReference getEndpointReference()
    {
        return m_epr;
    }

    public void setEndpointReference( EndpointReference epr )
    {
        m_epr = epr;
    }

    public void setNotificationConsumer( NotificationConsumer notificationConsumer )
    {
        m_notificationConsumer = notificationConsumer;
    }

    public void setNotificationProducer( NotificationProducer notificationProducer )
    {
        m_notificationProducer = notificationProducer;
    }

    public void addTerminationListener( ResourceTerminationListener resourceTerminationListener )
    {
        m_terminationListeners.add( resourceTerminationListener );
    }

    public void setPolicy( Object policy )
    {
        m_policy = policy;
    }

    public void setPrecondition( QueryExpression precondition )
    {
        m_precondition = precondition;
    }

    public void setSelector( QueryExpression selector )
    {
        m_selector = selector;
    }

    public void setUseNotify( boolean useNotify )
    {
        m_useNotify = useNotify;
    }
}
