/*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.ws.notification.topics.expression;

import org.apache.ws.notification.topics.expression.InvalidTopicExpressionException;
import org.apache.ws.notification.topics.expression.TopicExpressionException;
import org.apache.ws.notification.topics.expression.TopicExpressionResolutionException;
import org.apache.ws.notification.topics.expression.TopicPathDialectUnknownException;
import org.apache.ws.notification.topics.expression.InvalidTopicExpressionException;
import org.apache.ws.notification.topics.expression.TopicExpressionResolutionException;
import org.apache.ws.notification.topics.expression.TopicExpression;
import org.apache.ws.notification.topics.Topic;
import org.apache.ws.notification.topics.TopicSpaceSet;

/**
 * A <code>TopicExpressionEvaluator</code> is used to implement a topic expression evaluation against a
 * <code>TopicSpaceSet</code>. An evaluator can be registered with a <code>TopicExpressionEngine</code>,
 * which in turn calls the evaluator when a matching expression is found.
 */
public interface TopicExpressionEvaluator
{

    /**
     * Evaluates the expression over a TopicSpace and returns the set of matching topics.
     *
     * @param topicSpaceSet topic list associated with the service/resource
     * @param expression object passed by client representing the topic expression
     * @return the set of topics that matched the specified expression
     *
     * @throws TopicPathDialectUnknownException
     *                                  if the topic expression dialect is not supported
     * @throws TopicExpressionResolutionException
     *                                  if the expression could not be evaluated
     * @throws org.apache.ws.notification.topics.expression.InvalidTopicExpressionException
     *                                  if the topic expression is invalid
     * @throws TopicExpressionException if any other error occurs
     */
    Topic[] evaluate( TopicSpaceSet topicSpaceSet, TopicExpression expression )
            throws TopicPathDialectUnknownException,
            TopicExpressionResolutionException,
            InvalidTopicExpressionException,
            TopicExpressionException;

    /**
     * Gets the URIs for the dialects that this evaluator can handle
     *
     * @return array of URIs supported by this evaluator
     */
    String[] getDialects();

}
