/*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.ws.notification.topics.impl;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.notification.topics.Topic;
import org.apache.ws.notification.topics.TopicSpace;
import org.apache.ws.notification.topics.expression.TopicExpression;
import org.apache.ws.notification.topics.TopicListener;
import org.apache.ws.notification.topics.expression.TopicExpression;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;

/**
 * Simple in-memory implementation of the Topic interface
 */
public class TopicImpl implements Topic, TopicListener
{

    protected Map m_subTopics;
    protected TopicExpression m_reference;
    protected Collection m_listeners;
    protected String m_name;
    protected Object m_currentMsg;
    protected List m_topicPath; // ordered set of QNames
    protected TopicSpace m_topicSpace;
    private Topic m_parent;

    private static final Log LOG = LogFactory.getLog( TopicImpl.class.getName() );

    // TODO: throw more specific exceptions
    public Topic addTopic( Topic topic ) throws Exception
    {
        if ( m_reference != null )
        {
            throw new Exception( "addingSubTopicToReference" );
        }

        m_subTopics.put( topic.getName(), topic );

        List topicPath = new LinkedList();

        topicPath.addAll( m_topicPath );
        topicPath.add( topic.getName() );

        topic.setTopicPath( topicPath );
        topic.addTopicListener( this );

        topic.setTopicSpace(m_topicSpace);
        ((TopicImpl)topic).setParent( this );

        this.topicAdded( topic );
        return topic;
    }

    public Topic addTopic( String name ) throws Exception
    {
        return addTopic( new TopicImpl( name ) );
    }

    public void removeTopic( String name )
    {
        m_subTopics.remove( name );
    }

    public synchronized void addTopicListener( TopicListener listener )
    {
        m_listeners.add( listener );
    }

    public String getName()
    {
        return m_name;
    }

    public Topic getTopic( String topicName )
    {
        return (Topic) m_subTopics.get( topicName );
    }

    public boolean containsTopic( String name )
    {
        return m_subTopics.containsKey( name );
    }

    /**
     * Get the topic expression for the topic(s) that this topic references.
     *
     * @return The topic expression that this topic reference or null if this topic is not a reference.
     */
    public TopicExpression getTopicReference()
    {
        return m_reference;
    }

    public boolean isReference()
    {
        return m_reference != null;
    }

    public Iterator topicIterator()
    {
        return m_subTopics.values().iterator();
    }

    public void publish( Object msg ) throws Exception
    {
        m_currentMsg = msg;
        if ( LOG.isDebugEnabled() )
        {
            LOG.debug( "Publish called on topic " +
                    m_name + " with message " + msg );
        }
        this.topicChanged( this );
    }

    public Object getCurrentMessage()
    {
        return m_currentMsg;
    }

    public void removeTopic( Topic topic )
    {
        m_subTopics.remove( topic.getName() );
        topic.removeTopicListener( this );
        this.topicRemoved( topic );
    }

    public synchronized void removeTopicListener( TopicListener listener )
    {
        this.m_listeners.remove( listener );
    }

    public synchronized Iterator topicListenerIterator()
    {
        return this.m_listeners.iterator();
    }

    public void setTopicReference( TopicExpression topicPath )
    {
        this.m_reference = topicPath;
    }

    public void setTopicPath( List topicPath )
    {
        m_topicPath = topicPath;
    }

    public List getTopicPath()
    {
        return m_topicPath;
    }

    /**
     * Create a topic with the given name
     *
     * @param name The name of the created topic
     */
    public TopicImpl( String name )
    {
        this( new HashMap(), null, new Vector(), name, null, null );
    }

    /**
     * Create a topic with the given parameters
     *
     * @param subTopics A map of child topics
     * @param reference A topic expression (only used if this is a topic alias)
     * @param listeners A collection of topic listeners
     * @param name      The name of this topic
     * @param current   The current value of this topic
     * @param topicPath The concrete topic path of this topic
     */
    public TopicImpl( Map subTopics, TopicExpression reference,
                      Collection listeners, String name,
                      Object current, List topicPath )
    {
        this.m_subTopics = subTopics;
        this.m_reference = reference;
        this.m_listeners = listeners;
        this.m_name = name;
        this.m_currentMsg = current;
        if(topicPath==null){
            this.m_topicPath = new java.util.LinkedList();
            m_topicPath.add(name);
        }
    }

    public synchronized void topicAdded( Topic topic )
    {
        Iterator listenerIterator = this.m_listeners.iterator();
        TopicListener listener;

        while ( listenerIterator.hasNext() )
        {
            listener = (TopicListener) listenerIterator.next();
            listener.topicAdded( topic );
        }
    }

    public synchronized void topicChanged( Topic topic )
    {
        Iterator listenerIterator = this.m_listeners.iterator();
        TopicListener listener;
        while ( listenerIterator.hasNext() )
        {
            listener = (TopicListener) listenerIterator.next();
            listener.topicChanged( this );
        }
    }

    public synchronized void topicRemoved( Topic topic )
    {
        Iterator listenerIterator = this.m_listeners.iterator();
        TopicListener listener;
        while ( listenerIterator.hasNext() )
        {
            listener = (TopicListener) listenerIterator.next();
            listener.topicRemoved( topic );
        }
    }

    void setParent( Topic parent )
    {
        m_parent = parent;
    }

    public Topic getParent()
    {
        return m_parent;
    }
    
    public String toString(){
        // TODO: make this print the full concrete path of this Topic instead of just its name
        return getName();
    }
    
    public void setTopicSpace(TopicSpace topicSpace) {
        m_topicSpace = topicSpace;
    }
    
    public TopicSpace getTopicSpace() {
        return m_topicSpace;
    }

}
