/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.notification.topics.util;

import org.apache.ws.notification.base.v2004_06.porttype.NotificationProducerPortType;
import org.apache.ws.notification.topics.ResourcePropertyValueChangeTopic;
import org.apache.ws.notification.topics.Topic;
import org.apache.ws.notification.topics.TopicSpace;
import org.apache.ws.notification.topics.TopicSpaceSet;
import org.apache.ws.notification.topics.impl.ResourcePropertyValueChangeTopicImpl;
import org.apache.ws.notification.topics.impl.ResourceTerminationTopicImpl;
import org.apache.ws.notification.topics.expression.impl.TopicExpressionEngineImpl;
import org.apache.ws.notification.topics.impl.TopicSpaceImpl;
import org.apache.ws.notification.topics.v2004_06.TopicsConstants;
import org.apache.ws.resource.Resource;
import org.apache.ws.resource.lifetime.ResourceTerminationListener;
import org.apache.ws.resource.properties.NamespaceVersionHolder;
import org.apache.ws.resource.properties.ResourceProperty;
import org.apache.ws.resource.properties.ResourcePropertySet;
import org.apache.ws.util.XmlBeanUtils;
import org.oasisOpen.docs.wsn.x2004.x06.wsnWSBaseNotification12Draft01.FixedTopicSetDocument;
import org.oasisOpen.docs.wsn.x2004.x06.wsnWSBaseNotification12Draft01.TopicDocument;
import org.oasisOpen.docs.wsn.x2004.x06.wsnWSBaseNotification12Draft01.TopicExpressionDialectsDocument;
import org.oasisOpen.docs.wsn.x2004.x06.wsnWSBaseNotification12Draft01.TopicExpressionType;

import javax.xml.namespace.QName;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * TODO
 */
public class TopicUtils
{

    /**
     * Adds the topic for ResoruceTermination....there should be only one of these!
     *
     * @param topicSpaceSet
     * @return
     */
    public static Topic addResourceTerminationTopic(TopicSpaceSet topicSpaceSet, Resource resource, NamespaceVersionHolder namespaces)
            throws Exception
    {
        //there can be only 1 !
        String namespace = namespaces.getLifetimeXsdNamespace();        
        TopicSpace topicSpace = topicSpaceSet.getTopicSpace(namespace);
        if (topicSpace == null)
        {
            topicSpace = new TopicSpaceImpl(namespace);
            topicSpaceSet.addTopicSpace(topicSpace);
        }
        Topic topic;
        if (topicSpace.topicIterator().hasNext())
        {
            topic = (Topic) topicSpace.topicIterator().next();
        }
        else
        {
            topic = new ResourceTerminationTopicImpl(namespaces);
            resource.addTerminationListener((ResourceTerminationListener) topic);
            topicSpace.addTopic(topic);
        }
        return topic;
    }

    /**
     * Adds a topic for the specified property to the specified topic set.
     *
     * @param prop
     * @param topicSpaceSet
     * @return
     */
    public static Topic addResourcePropertyValueChangeTopic( ResourceProperty prop, TopicSpaceSet topicSpaceSet )
            throws Exception
    {
        if ( prop.getMetaData().isReadOnly() )
        {
            throw new IllegalArgumentException( "Property " + prop.getMetaData().getName() + " is readonly!" );
        }
        ResourcePropertyValueChangeTopic valueChangeTopic = new ResourcePropertyValueChangeTopicImpl( prop );
        prop.addChangeListener( valueChangeTopic );
        String topicNsURI = prop.getMetaData().getName().getNamespaceURI();
        TopicSpace topicSpace = topicSpaceSet.getTopicSpace( topicNsURI );
        if ( topicSpace == null )
        {
            topicSpace = new TopicSpaceImpl( topicNsURI );
            topicSpaceSet.addTopicSpace( topicSpace );
        }
        topicSpace.addTopic( valueChangeTopic );        
        return valueChangeTopic;
    }

    /**
     * Adds topics for all non-readonly properties from the specified property set to the specified topic set.
     *
     * @param propSet
     * @param topicSpaceSet
     * @return
     */
    public static Topic[] addResourcePropertyValueChangeTopics( ResourcePropertySet propSet, TopicSpaceSet topicSpaceSet )
            throws Exception
    {
        List topicList = new ArrayList();
        Iterator propIter = propSet.iterator();
        while ( propIter.hasNext() )
        {
            ResourceProperty prop = (ResourceProperty) propIter.next();
            if ( ! prop.getMetaData().isReadOnly() )
            {
                Topic topic = addResourcePropertyValueChangeTopic( prop, topicSpaceSet );
                topicList.add( topic );
            }
        }
        return (Topic[]) topicList.toArray( new Topic[0] );
    }

    /**
     * Initializes the Topic, TopicExpressionDialects and FixedTopicSet resource properties
     *
     * </br>
     * <strong>NOTE this MUST be called last in the Resource.init() method to ensure all topics get registered</strong>
     * </br>
     * FixedTopicSet will be set based on the value returned from the {@see TopicSpaceSet#isFixed()} method.
     * </br>
     * Topic will have all root topics in the TopicSpaceSet, set to Simple dialect AND all child topics set ot
     * Concrete dialect.
     * </br>
     *
     * TopicExpressionDialects will be set to the engine's known Topic Dialects acquired from the TopicExpressionEngine
     *
     * @param topicSpaceSet
     * @param propSet
     */
    public static void initNotificationProducerProperties(TopicSpaceSet topicSpaceSet, ResourcePropertySet propSet)
    {
        //setup the FixedTopicSet property
        ResourceProperty resourceProperty = propSet.get(NotificationProducerPortType.PROP_QNAME_FIXED_TOPIC_SET);
        FixedTopicSetDocument fixedTopicSetDocument = FixedTopicSetDocument.Factory.newInstance();
        fixedTopicSetDocument.setFixedTopicSet(topicSpaceSet.isFixed());
        resourceProperty.add(fixedTopicSetDocument);

        //setup the Topic property
        resourceProperty = propSet.get(NotificationProducerPortType.PROP_QNAME_TOPIC);
        TopicSpace[] topicSpaces = topicSpaceSet.getTopicSpaces();
        for (int i = 0; i < topicSpaces.length; i++)
        {
            TopicSpace topicSpace = topicSpaces[i];
            String targetNamespace = topicSpace.getTargetNamespace();
            Iterator iterator = topicSpace.topicIterator();
            while (iterator.hasNext())
            {
                Topic topic =  (Topic)iterator.next();
                String topicName = topic.getName();
                QName rootTopicQName = new QName(targetNamespace, topicName);
                addSimpleTopic(rootTopicQName, resourceProperty);
                Iterator subTopics = topic.topicIterator();
                while (subTopics.hasNext())
                {
                    Topic subTopic = (Topic) subTopics.next();
                    addConcreteTopics(rootTopicQName, subTopic, resourceProperty);
                }
            }
        }


        //setup the TopicExpressionDialects property
        resourceProperty = propSet.get(NotificationProducerPortType.PROP_QNAME_TOPIC_EXPRESSION_DIALECTS);
        String[] supportedDialects = TopicExpressionEngineImpl.getInstance().getSupportedDialects();
        //add a TopicExpressionDialectsDocument for each TopicExpressionDialect.
        for (int i = 0; i < supportedDialects.length; i++)
        {
            TopicExpressionDialectsDocument topicExpressionDialectsDocument = TopicExpressionDialectsDocument.Factory.newInstance();
            topicExpressionDialectsDocument.setTopicExpressionDialects(supportedDialects[i]);
            resourceProperty.add(topicExpressionDialectsDocument);
        }
    }

    private static void addSimpleTopic(QName rootTopicQName, ResourceProperty resourceProp)
    {
        TopicDocument topicDocument = TopicDocument.Factory.newInstance();
        TopicExpressionType topicExpressionType = topicDocument.addNewTopic();
        topicExpressionType.setDialect(TopicsConstants.TOPIC_EXPR_DIALECT_SIMPLE);
        XmlBeanUtils.setValueAsQName(topicExpressionType, rootTopicQName);
        resourceProp.add(topicDocument);
    }

    /**
     * Recursively adds Concrete Topics to the Topic Document
     *
     * @param rootTopicQName The QName of the parent topic.
     * @param subTopic  The current topic being added.
     * @param resourceProp  The ResourceProperty associated with the Topic being added to.
     */
    private static void addConcreteTopics(QName rootTopicQName, Topic subTopic, ResourceProperty resourceProp)
    {
        TopicDocument topicDocument = TopicDocument.Factory.newInstance();
        String topicName = subTopic.getName();
        QName concreteTopicName = new QName(rootTopicQName.getNamespaceURI(), rootTopicQName.getLocalPart() + "/" + topicName);
        TopicExpressionType topicExpressionType = topicDocument.addNewTopic();
        topicExpressionType.setDialect(TopicsConstants.TOPIC_EXPR_DIALECT_CONCRETE);
        XmlBeanUtils.setValueAsQName(topicExpressionType, concreteTopicName);
        resourceProp.add(topicDocument);
        Iterator iterator = subTopic.topicIterator();
        while (iterator.hasNext())
        {
            Topic topic = (Topic) iterator.next();
            addConcreteTopics(concreteTopicName, topic, resourceProp);
        }
    }

}
