/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.pubsub.emitter.file;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.pubsub.emitter.EmitterException;
import org.apache.ws.pubsub.emitter.EmitterTask;
import org.apache.ws.pubsub.i18n.Keys;
import org.apache.ws.pubsub.i18n.MessagesImpl;
import org.apache.ws.util.i18n.Messages;

import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPMessage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * An emitter task that stores a notification to a file specified in the FILE protocol URL.
 * If the file already exists, the new notification data will be appended to the end.
 * If the file is actually a directory, the notification filename will be generated.
 */
public class FileEmitterTask
   extends EmitterTask
{
   /** This emitter's protocol */
   public static final String PROTOCOL = "file";

   /** object used to obtain I18N messages */
   private static final Messages MSG = MessagesImpl.getInstance(  );

   /** object used to log messages */
   private static final Log LOG = LogFactory.getLog( FileEmitterTask.class );

   /**
    * Creates a new {@link FileEmitterTask} object.
    *
    * @see EmitterTask(SOAPMessage, URL)
    */
   public FileEmitterTask( SOAPMessage           notif,
                                       URL                   url)
   {
      super( notif, url );
   }

   /**
    * Stores the notification in a file specified by the destination URL.  No attempt to
    * lock the file or otherwise force synchronous writes to the file is made;
    * it is up to the subscriber to provide a file URI that will be suitable for
    * writing.  If the file is a directory, a file will be created in that directory.
    *
    * @see EmitterTask#emit()
    */
   protected void emit(  )
   throws EmitterException
   {
      LOG.debug( MSG.getMessage( Keys.EMITTING_TO_FILE,
                                 getDestinationUrl(  ) ) );

      try
      {
         File destFile = new File( getDestinationUrl(  ).getPath(  ) );

         if ( destFile.isDirectory(  ) )
         {
            destFile = getDefaultDestinationFile( destFile );
         }
         else
         {
            destFile.getParentFile(  ).mkdirs(  );
         }

         FileOutputStream fos          = new FileOutputStream( destFile, true );
         SOAPMessage      notification = getNotification(  );
         notification.writeTo( fos );
         fos.close(  );
      }
      catch ( MalformedURLException murle )
      {
         throw new EmitterException( MSG.getMessage( Keys.EX_FAILED_TO_STORE_NOTIFICATION_TO_FILE_BAD_URL,
                                                     getDestinationUrl(  ) ), murle );
      }
      catch ( IOException ioe )
      {
         throw new EmitterException( MSG.getMessage( Keys.EX_FAILED_TO_STORE_NOTIFICATION_TO_FILE,
                                                     getDestinationUrl(  ) ), ioe );
      }
      catch ( SOAPException e )         
      {
         throw new EmitterException( MSG.getMessage( Keys.EX_FAILED_TO_STORE_NOTIFICATION_TO_FILE,
                                                     getDestinationUrl(  ) ), e );
      }

      return;
   }

   /**
    * Determines a default file to use when writing the notification to the directory.  The file will be created
    * when this method returns.
    *
    * @param dir the directory
    *
    * @return the file where the notification will be stored
    * @throws IOException if failed to create the temp file
    */
   private File getDefaultDestinationFile( File dir )
   throws IOException
   {
      return File.createTempFile( "notification-", ".txt", dir );
   }
}