/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.util.thread;

import EDU.oswego.cs.dl.util.concurrent.ThreadFactory;

/**
 * Subclass of {@link java.lang.Thread} that provides commonly named threads for
 * use inside a particular software application.
 * <p/>
 * The main purpose of this class is to provide all threads
 * running in the software with a consistent naming convention.  When instantiating
 * a <code>NamedThread</code>, if a name is provided, that name will be used as the
 * base name for the thread.  The base name will be decorated with additional information
 * to make the thread name unique.
 */
public class NamedThread
   extends Thread
{
   /** all thread names will have this string as a prefix */
   public static final String THREAD_NAME_PREFIX = "hermes.thread.";

   /** number of named threads that have been created */
   private static long s_threadCount = 0L;

   /** base name used to create this thread; or <code>null</code> if no name was specified */
   private final String m_baseName;

   /**
    * Creates the thread with an empty base name.
    *
    * @see java.lang.Thread#Thread(java.lang.String)
    */
   public NamedThread(  )
   {
      super( createThreadName( null ) );
      m_baseName = null;
   }

   /**
    * Creates the thread with an empty base name.
    *
    * @see java.lang.Thread#Thread(java.lang.Runnable)
    */
   public NamedThread( Runnable target )
   {
      super( target,
             createThreadName( null ) );
      m_baseName = null;
   }

   /**
    * Creates the thread with an empty base name.
    *
    * @see java.lang.Thread#Thread(java.lang.ThreadGroup, java.lang.Runnable)
    */
   public NamedThread( ThreadGroup group,
                       Runnable    target )
   {
      super( group, target,
             createThreadName( null ) );
      m_baseName = null;
   }

   /**
    * Creates the thread with the given name to be used as the base name.
    *
    * @see java.lang.Thread#Thread(java.lang.String)
    */
   public NamedThread( String name )
   {
      super( createThreadName( name ) );
      m_baseName = name;
   }

   /**
    * Creates the thread with the given name to be used as the base name.
    *
    * @see java.lang.Thread#Thread(java.lang.ThreadGroup, java.lang.String)
    */
   public NamedThread( ThreadGroup group,
                       String      name )
   {
      super( group,
             createThreadName( name ) );
      m_baseName = name;
   }

   /**
    * Creates the thread with the given name to be used as the base name.
    *
    * @see java.lang.Thread#Thread(java.lang.Runnable, java.lang.String)
    */
   public NamedThread( Runnable target,
                       String   name )
   {
      super( target,
             createThreadName( name ) );
      m_baseName = name;
   }

   /**
    * Creates the thread with the given name to be used as the base name.
    *
    * @see java.lang.Thread#Thread(java.lang.ThreadGroup, java.lang.Runnable, java.lang.String)
    */
   public NamedThread( ThreadGroup group,
                       Runnable    target,
                       String      name )
   {
      super( group, target,
             createThreadName( name ) );
      m_baseName = name;
   }

   /**
    * Returns the base name used to create the thread.  To get the actual name
    * of the thread, use {@link java.lang.Thread#getName()}.  If no name was
    * passed into a constructor, or if null was passed, then null is returned.
    *
    * @return base name of the thread (the name passed into the constructor)
    */
   public String getBaseName(  )
   {
      return m_baseName;
   }

   /**
    * Builds a thread name that is unique to the JVM.
    * The base name provided will be used as part of the thread
    * name to further describe the thread's purpose.  If the base name is
    * null, it will not be used.
    *
    * @param baseName thread's base name
    *
    * @return String a String that can be used as a thread name
    */
   protected static String createThreadName( String baseName )
   {
      String name;
      long   thread_num;

      synchronized ( THREAD_NAME_PREFIX )
      { // really we want to synchronize on m_threadCount, but its a primitive

         // so just use THREAD_NAME_PREFIX as a monitor lock
         thread_num = s_threadCount++;
      }

      name = THREAD_NAME_PREFIX + thread_num;

      if ( baseName != null )
      {
         name += ( "." + baseName );
      }

      return name;
   }

   /**
    * A thread factory for use with the Concurrent API.  This factory creates threads
    * of type {@link NamedThread} and provides the ability to name the threads via the
    * base name facility of {@link NamedThread}.  Threads created by this
    * factory may be daemon threads if desired
    * (see {@link ConcurrentThreadFactory#ConcurrentThreadFactory(String, boolean)}).
    */
   public static class ConcurrentThreadFactory
      implements ThreadFactory
   {
      /** the base name this factory will use for all of its threads */
      private String m_factoryBaseName;

      /** determines if all threads created by this factory are daemon threads or not */
      private boolean m_isDaemon;

      /**
       * Creates a factory that does not define a base name for the user threads it will create.
       */
      public ConcurrentThreadFactory(  )
      {
         this( null, false );
      }

      /**
       * Creates a factory that defines a base name for the user threads it will create.
       *
       * @param baseName the base name used for all threads created by this factory (may be <code>null</code>)
       *
       * @see NamedThread#getBaseName()
       */
      public ConcurrentThreadFactory( String baseName )
      {
         this( baseName, false );
      }

      /**
       * Creates a factory that does not define a base name for the threads it will create but explicitly
       * defines whether or not the threads will be daemon or user threads.
       *
       * @param isDaemon if <code>true</code> then all threads created by this factory will be daemon threads.
       */
      public ConcurrentThreadFactory( boolean isDaemon )
      {
         this( null, isDaemon );
      }

      /**
       * Creates a factory that defines the new threads' base names and whether or not the threads to be
       * created will be daemon or user threads.
       *
       * @param baseName the base name used for all threads created by this factory (may be <code>null</code>)
       * @param isDaemon if <code>true</code> then all threads created by this factory will be daemon threads.
       */
      public ConcurrentThreadFactory( String  baseName,
                                      boolean isDaemon )
      {
         m_factoryBaseName    = baseName;
         m_isDaemon           = isDaemon;
      }

      /**
       * Creates a {@link NamedThread} using this factory's base name as the new thread's base name.
       *
       * @see ThreadFactory#newThread(java.lang.Runnable)
       * @see NamedThread#NamedThread(Runnable, String)
       * @see NamedThread#getBaseName()
       */
      public Thread newThread( Runnable command )
      {
         NamedThread retNewThread = new NamedThread( command, m_factoryBaseName );

         retNewThread.setDaemon( m_isDaemon );

         return retNewThread;
      }
   }
}
