/*
 * Copyright  2004 The Apache Software Foundation.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package org.apache.ws.util.uuid.impl;

import org.apache.ws.util.uuid.UuidGenerator;

import javax.xml.rpc.JAXRPCException;
import java.lang.reflect.Method;

/**
 * A {@link org.apache.ws.util.uuid.UuidGenerator} implementation that uses JUG, an opensource Java UUID generation
 * library. Reflection is utilized to avoid a compile-time dependency on jug.jar.
 *
 * @author Ian P. Springer <ian_springer@hp.com>
 */
public class JugUuidGenerator implements UuidGenerator
{

    private static final String JUG_UUIDGENERATOR_CLASSNAME = "org.doomdark.uuid.UUIDGenerator";
    private static final String JUG_WEBSITE_URL = "http://www.doomdark.org/doomdark/proj/jug/";

    private static Method generateUUIdMethod;
    private static Method newInstanceMethod;

    static
    {
        Class jugUuidGenClass = null;
        try
        {
            jugUuidGenClass = Class.forName( JUG_UUIDGENERATOR_CLASSNAME );
        }
        catch ( ClassNotFoundException cnfe )
        {
            throw new RuntimeException(
                    "Java UUID Generator (JUG) " + JUG_UUIDGENERATOR_CLASSNAME +
                    " not found. Please add jug.jar, from " +
                    JUG_WEBSITE_URL +
                    ", to your classpath and restart." );
        }
        try
        {
            newInstanceMethod = jugUuidGenClass.getMethod( "getInstance", new Class[0] );
            generateUUIdMethod = jugUuidGenClass.getMethod( "generateTimeBasedUUID", new Class[0] );
        }
        catch ( Exception e )
        {
            throw new RuntimeException(
                    "Fatal error initializing Java UUID Generator (JUG) " + JUG_UUIDGENERATOR_CLASSNAME + "." );
        }
    }

    private ThreadLocal JUG_UUIDGENERATOR =
            new ThreadLocal()
            {
                protected synchronized Object initialValue()
                {
                    try
                    {
                        return newInstanceMethod.invoke( null, new Object[0] );
                    }
                    catch ( Exception e )
                    {
                        throw new RuntimeException( e );
                    }
                }
            };

    public String generateUuid()
    {
        try
        {
            return generateUUIdMethod.invoke( JUG_UUIDGENERATOR.get(), new Object[0] ).toString();
        }
        catch ( Exception e )
        {
            throw new JAXRPCException( e );
        }
    }

}
